
/*

    File: ftp.c

    Copyright (C) 1998,1999  Wolfgang Zekoll  <wzk@quietsche-entchen.de>
    Copyright (c) 2000,2002,2003  Andreas Schoenberg  <asg@pce.de>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include <time.h>
#include <signal.h>
#include <wait.h>
#include <ctype.h>
#include <pwd.h>
#include <shadow.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <sys/time.h>

#include "ftp.h"
#include "ip-lib.h"
#include "lib.h"


	/*
	 * Handlertypen
	 */
	 
#define	H_NONE			0
#define	H_PRE			1
#define	H_POST			2


typedef struct _ftpcmd {
    char	name[20];
    int		par, ispath;
    int		useccp, usehandler;
    int		log;
    } ftpcmd_t;

ftpcmd_t cmdtab[] = {

	/*
	 * Einfache FTP Kommandos.
	 */

    "MODE",  1, 0, 0, H_NONE, 0,
    "TYPE",  1, 0, 0, H_NONE, 0,
    "SYST",  0, 0, 0, H_NONE, 0,
    "NOOP",  0, 0, 0, H_NONE, 0,
    "PORT",  1, 0, 0, H_NONE, 0,
    "PASV",  0, 0, 0, H_NONE, 0,
    "ABOR",  0, 0, 0, H_NONE, 1,
    "QUIT",  0, 0, 0, H_NONE, 1,
    
    "CDUP",  1, 1, 1, H_POST, 1,
    "CWD",   1, 1, 1, H_POST, 1,
    "PWD",   0, 0, 0, H_NONE, 0,
    "MKD",   1, 1, 1, H_POST, 1,
    "RMD",   1, 1, 1, H_POST, 1,
    
    "LIST",  1, 1, 1, H_POST, 0,
    "NLST",  1, 1, 1, H_POST, 0,

    "REST",  1, 0, 0, H_NONE, 0,
    "RETR",  1, 1, 1, H_NONE, 0,	/* Handler wird manuell gestartet. */
    "STOR",  1, 1, 1, H_NONE, 0,	/* Die nachfolgenden Upload Handler auch. */
    "STOU",  0, 0, 1, H_NONE, 0,
    "APPE",  1, 1, 1, H_NONE, 0,

    "MDTM",  1, 1, 1, H_NONE, 0,
    "SIZE",  1, 1, 1, H_NONE, 0,

    "DELE",  1, 1, 1, H_POST, 0,
    "RNFR",  1, 1, 1, H_POST, 1,
    "RNTO",  1, 1, 1, H_POST, 1,
    
    "SITE",  1, 1, 0, H_NONE, 0,

	/*
	 * Nicht unterstuetzte Kommandos
	 */

    "STAT", -1, 0, 1, H_NONE, 0,
    "ACCT", -1, 0, 0, H_NONE, 0,
    "REIN", -1, 0, 0, H_NONE, 0,
    "SMNT", -1, 0, 0, H_NONE, 0,
    "STRU", -1, 0, 0, H_NONE, 0,
    "ALLO", -1, 0, 0, H_NONE, 0,
    "HELP", -1, 0, 0, H_NONE, 0,

    "", 0, 0, 0, 0
    };


unsigned get_interface_info(int pfd, char *ip, int max)
{
	int	size;
	unsigned int port;
	struct sockaddr_in saddr;

	size = sizeof(saddr);
	if (getsockname(pfd, (struct sockaddr *) &saddr, &size) < 0) {
		syslog(LOG_NOTICE, "-ERR: can't get interface info: %m");
		exit (-1);
		}

	copy_string(ip, (char *) inet_ntoa(saddr.sin_addr), max);
	port = ntohs(saddr.sin_port);

	return (port);
}

int get_client_info(ftp_t *x, int pfd)
{
	int	size;
	struct sockaddr_in saddr;
	struct in_addr *addr;
	struct hostent *hostp = NULL;

	*x->client = 0;
	size = sizeof(saddr);
	if (getpeername(pfd, (struct sockaddr *) &saddr, &size) < 0 )
		return (-1);
		
	copy_string(x->client_ip, (char *) inet_ntoa(saddr.sin_addr), sizeof(x->client_ip));
	addr = &saddr.sin_addr,
	hostp = gethostbyaddr((char *) addr,
			sizeof (saddr.sin_addr.s_addr), AF_INET);

	hostp = NULL;
	if (hostp == NULL) {
		strcpy(x->client, x->client_ip);
		return (0);
		}

	strcpy(x->client, hostp->h_name);
	strlwr(x->client);

	return (0);
}


	/*
	 * Basic I/O functions
	 */

int asc2unix(char *data, char *buffer, int len)
{
	int	c, i, k;

	buffer[len] = 0;
	k = 0;
	for (i=0; i<len; i++) {
		if ((c = buffer[i]) == '\r'  &&  buffer[i+1] == '\n')
			continue;

		data[k++] = c;
		}

	data[k] = 0;
	return (k);
}

int unix2asc(char *data, char *buffer, int len)
{
	int	c, i, k;

	buffer[len] = 0;
	k = 0;
	for (i=0; i<len; i++) {
		if ((c = buffer[i]) == '\r'  &&  buffer[i+1] == '\n')
			continue;
		else if (c == '\n')
			data[k++] = '\r';

		data[k++] = c;
		}

	data[k] = 0;
	return (k);
}


int close_ch(ftp_t *x, dtc_t *ch)
{
	int	rc;

	if (ch->local >= 0) {
		close(ch->local);
		if (ch->localpipe)
			wait(&rc);
		}

	if (ch->data >= 0)
		close (ch->data);

	ch->local     = -1;
	ch->data      = -1;
	ch->state     = 0;
	ch->operation = 0;

	return (0);
}

int connect_ch(ftp_t *x, dtc_t *ch, int op, int fd, int ispipe, int type,
		char *command, char *filename)
{
	ch->operation = op;
	ch->local     = fd;
	ch->localpipe = ispipe;
	ch->type      = type;

	if (ch->mode != MODE_PASSIVE) {
		if ((ch->data = openip(ch->client.ipnum, ch->client.port, x->interface, 0)) < 0) {
			syslog(LOG_NOTICE, "-ERR: can't connect to client: %m");
			exit (1);
			}
		}
	else {
		int	sock, adrlen;
		char	remote[80];
		struct sockaddr_in adr;

		adrlen = sizeof(struct sockaddr);
		sock = accept(ch->data, (struct sockaddr *) &adr, &adrlen);
		if (debug != 0)
			fprintf (stderr, "accept() on socket\n");

		if (sock < 0) {
			syslog(LOG_NOTICE, "-ERR: accept error: %m");
			exit (1);
			}


		copy_string(remote, inet_ntoa(adr.sin_addr), sizeof(remote));
		if (debug)
			fprintf (stderr, "connection from %s\n", remote);

		/*
		 * Gegenstelle ueberpruefen.
		 */

		if (strcmp(x->client_ip, remote) != 0) {
			if (x->config->ignore_anyremote == 0)
				syslog(LOG_NOTICE, "-ERR: unexpected connect: %s, expected= %s", remote, x->client_ip);

			if (x->config->allow_anyremote == 0)
				exit (1);
			}


		close(ch->data);
		ch->data = sock;
		}


	if (debug)
		fprintf (stderr, "local= %d\n", ch->local);

	if (ch->operation == OP_GET) {
		ch->active = ch->local;
		ch->other  = ch->data;
		}
	else if (x->ch.operation == OP_PUT) {
		ch->active = ch->data;
		ch->other  = ch->local;
		}
	else {
		syslog(LOG_NOTICE, "-ERR: transfer operation error");
		exit (1);
		}

	ch->bytes = 0;
	ch->state = PORT_CONNECTED;
	
	copy_string(ch->command, command, sizeof(ch->command));
	copy_string(ch->filename, filename, sizeof(ch->filename));

	return (0);
}

int getc_fd(ftp_t *x, int fd)
{
	int	c;

	if (x->bio.here >= x->bio.len) {
		int	rc, max, bytes;
		struct timeval tov;
		fd_set	available, fdset;

		x->bio.len = x->bio.here = 0;

		bytes = 0;
		while (1) {
			FD_ZERO(&fdset);
			FD_SET(fd, &fdset);
			max = fd;

			if (x->ch.operation == 0)
				/* nichts */ ;
			else if (x->ch.state == PORT_CONNECTED) {
				FD_SET(x->ch.active, &fdset);
				if (x->ch.active > max)
					max = x->ch.active;
				}

			memmove(&available, &fdset, sizeof(fd_set));
			tov.tv_sec  = x->config->timeout;
			tov.tv_usec = 0;

			if (debug != 0)
				fprintf (stderr, "waiting for data, max= %d\n", max);

			rc = select(max + 1, &available, (fd_set *) NULL, (fd_set *) NULL, &tov);
			if (rc < 0) {
				syslog(LOG_NOTICE, "select() error: %m\n");
				break;
				}
			else if (rc == 0) {
				syslog(LOG_NOTICE, "connection timed out: client= %s",
					x->client);
				return (-1);
				}

			if (FD_ISSET(fd, &available)) {
				if ((bytes = read(fd, x->bio.buffer, sizeof(x->bio.buffer))) <= 0)
					return (-1);

				break;
				}
			else if (FD_ISSET(x->ch.active, &available)) {
				if (x->ch.state == PORT_CONNECTED) {
					char	*data, buffer[FTPBUFLEN], translated[(2 * FTPBUFLEN)];

					bytes = read(x->ch.active, buffer, sizeof(buffer));
					if (debug != 0)
						fprintf (stderr, "read %d bytes\n", bytes);

					if (bytes > 0) {
						data = buffer;
						if (x->ch.type == TYPE_ASC) {
							if (x->ch.operation == OP_GET)
								bytes = unix2asc(data = translated, buffer, bytes);
							else
								bytes = asc2unix(data = translated, buffer, bytes);
							}
						
						bytes = write(x->ch.other, data, bytes);
						if (debug != 0)
							fprintf (stderr, "wrote %d bytes\n", bytes);

						x->ch.bytes += bytes;
						}
					else {
						if (debug)
							fprintf (stderr, "closing data connection\n");

						close_ch(x, &x->ch);
						FD_ZERO(&fdset);
						FD_SET(fd, &fdset);
						max = fd;

						return (1);
						}
					}
				}
			}

		x->bio.len  = bytes;
		x->bio.here = 0;
		}

	if (x->bio.here >= x->bio.len)
		return (-1);

	c = (unsigned char) x->bio.buffer[x->bio.here++];
	return (c);
}

char *readline_fd(ftp_t *x, int fd, char *line, int size)
{
	int	c, k;

	if (fd != x->bio.lastfd) {
		x->bio.len  = 0;
		x->bio.here = 0;
		}
		
	x->bio.lastfd = fd;
	*line = 0;
	size = size - 2;

	c = getc_fd(x, fd);
	if (c < 0)
		return (NULL);
	else if (c == 1) {
		strcpy(line, "\001");
		return (line);
		}

	k = 0;
	while (c > 0  &&  c != '\n'  &&  c != 0) {
		if (k < size)
			line[k++] = c;

		c = getc_fd(x, fd);
		}

	line[k] = 0;
	noctrl(line);

				/* Why did I code this? */
	k = 0;
	while ((c = (unsigned char ) line[k]) != 0  &&  c > 126)
		k++;

	if (k > 0)
		copy_string(line, &line[k], size);

	return (line);
}


char *cfgets(ftp_t *x, char *line, int size)
{
	char	*p;

	*line = 0;
	if ((p = readline_fd(x, 0, line, size)) == NULL)
		return (NULL);
	else if (debug != 0)
		fprintf (stderr, "CLI >>>: %s\n", p);

	return (line);
}

int cfputs(ftp_t *x, char *line)
{
	int	bytes;
	char	buffer[310];

	if (debug)
		fprintf (stderr, ">>> CLI: %s\n", line);

	snprintf (buffer, sizeof(buffer) - 2, "%s\r\n", line);
	if ((bytes = write(1, buffer, strlen(buffer))) <= 0) {
		syslog(LOG_NOTICE, "-ERR: wrote %d/%d bytes to client, error= %m", bytes, strlen(line));
		exit (1);
		}
		
	return (0);
}

int cfpute(ftp_t *x, char *status)
{
	char	*error, line[200];

	error = sys_errlist[errno];
	snprintf (line, sizeof(line) - 2, "%s %s", status, error);
	strlwr(line);
	cfputs(x, line);

	return (0);
}




int doquit(ftp_t *x)
{
	cfputs(x, "221 goodbye");
	syslog(LOG_NOTICE, "QUIT");
	call_handler(x, "+QUIT", "", 0);
	
	return (0);
}


char *_getipnum(char *line, char **here, char *ip, int size)
{
	int	c, i, k;

	copy_string(ip, line, size);
	k = 0;
	for (i=0; (c = ip[i]) != 0; i++) {
		if (c == ',') {
			if (k < 3) {
				ip[i] = '.';
				k++;
				}
			else {
				ip[i++] = 0;
				break;
				}
			}
		}

	if (here != NULL)
		*here = &line[i];

	return (ip);
}

unsigned long _getport(char *line, char **here)
{
	unsigned long port;
	char	*p;

	p = line;
	port = strtoul(p, &p, 10);
	if (*p != ',')
		return (0);

	p++;
	port = (port << 8) + strtoul(p, &p, 10);
	if (here != NULL)
		*here = p;

	return (port);
}

int doport(ftp_t *x, char *command, char *par)
{
	char	*p;
	dtc_t	*ch;

	ch = &x->ch;
	if (ch->state != 0) {
		syslog(LOG_NOTICE, "PORT with already active port");
		close_ch(x, ch);
		}

	_getipnum(par, &p, ch->client.ipnum, sizeof(ch->client.ipnum));
	ch->client.port = _getport(p, &p);
	if (debug)
		fprintf (stderr, "client listens on %s:%u\n", ch->client.ipnum, ch->client.port);

	if (strcmp(ch->client.ipnum, x->client_ip) != 0) {
		if (x->config->ignore_anyremote == 0)
			syslog(LOG_NOTICE, "PORT %s:%u not on client %s", ch->client.ipnum, ch->client.port, x->interface);
		
		if (x->config->allow_anyremote == 0)
			exit (1);
		}

	cfputs(x, "200 ok");
	ch->local  = -1;

	ch->mode   = MODE_PORT;
	ch->state  = PORT_LISTEN;

	*ch->command = 0;
	return (0);
}

int dopasv(ftp_t *x)
{
	int	c;
	char	*p, line[200];
	dtc_t	*ch;

	ch = &x->ch;
	if (ch->state != 0) {
		syslog(LOG_NOTICE, "PASV with already active port");
		close_ch(x, ch);
		}

	copy_string(ch->pasv.ipnum, x->interface, sizeof(ch->pasv.ipnum));
	ch->data = bind_to_port(ch->pasv.ipnum, 0);
	ch->pasv.port = get_interface_info(ch->data, line, sizeof(line));
	if (debug)
		fprintf (stderr, "listening on %s:%u\n", ch->pasv.ipnum, ch->pasv.port);

	snprintf (line, sizeof(line) - 2, "227 ok, entering passive mode (%s,%u,%u)",
			ch->pasv.ipnum,
			ch->pasv.port >> 8, ch->pasv.port & 0xFF);
	for (p=line; (c = *p) != 0; p++) {
		if (c == '.')
			*p = ',';
		}

	cfputs(x, line);
	ch->mode  = MODE_PASSIVE;
	ch->state = PORT_LISTEN;

	*ch->command = 0;
	ch->operation = 0;

	return (0);
}



int setvar(ftp_t *x, char *var, char *value)
{
	char	varname[200];

	snprintf (varname, sizeof(varname) - 2, "%s%s", x->config->varname, var);
	setenv(varname, value != NULL? value: "", 1);

	return (0);
}

int setuserenv(ftp_t *x)
{
	unsigned long now;
	char	val[200];

	setvar(x, "INTERFACE", x->interface);
	snprintf (val, sizeof(val) - 2, "%u", x->port);
	setvar(x, "PORT", val);

	now = time(NULL);
	snprintf (val, sizeof(val) - 2, "%lu", now);
	setvar(x, "STARTED", val);

	snprintf (val, sizeof(val) - 2, "%d", getpid());
	setvar(x, "SERVERPID", val);

	setvar(x, "CLIENT", x->client_ip);
	setvar(x, "CLIENTNAME", x->client);

	setvar(x, "ROOTDIR", x->rootdir);
	setvar(x, "USER", x->username);
	setvar(x, "USERNAME", x->username);
	setvar(x, "HOME", x->home);

	return (0);
}

int set_variables(ftp_t *x, char *cmd, char *par, int ispath)
{
	setvar(x, "COMMAND", cmd == NULL? "": cmd);
	setvar(x, "SYSPATH", par == NULL? "": par);

	if (par != NULL  &&  *par != 0  &&  ispath != 0) {
		char	string[200];

		/*
		 * Wenn par gesetzt ist und ein ispath-Parameter, dann ist
		 * der Wert der absolute Pfad auf die Datei (bzw. Verzeichnis).
		 */

		setvar(x, "FILENAME", getfile(par, string, sizeof(string)));
		setvar(x, "SYSDIR", getdir(par, string, sizeof(string)));
		setvar(x, "FTPDIR", get_ftppath(x, string, sizeof(string)));
		}
	else if (strcmp(cmd, "STOU") == 0) {
		char	cwd[200];

		/*
		 * STOU hat zwar keinen Parameter, aber ein aktuelles Verzsichnis
		 * Das setzen wir, damit das ccp und der uniquename-handler
		 * entscheiden koennen, ob STOU erlaubt ist oder nicht.
		 *
		 * Bei den command handler Aufrufen ist der Dateiparemeter korrekt
		 * angegeben, so dass kein Workaround noetig ist.
		 */

		getcwd(cwd, sizeof(cwd));
		setvar(x, "SYSDIR", cwd);
		setvar(x, "FTPDIR", get_ftppath(x, cwd, sizeof(cwd)));
		}

	return (0);
}

int run_ccp(ftp_t *x, char *cmd, char *par, int ispath)
{
	int	rc, pid, pfd[2], lfd[2];
	char	message[200], line[300];

	/*
	 * Wenn kein ccp angegeben ist ist alles erlaubt.
	 */

	if (*x->config->ccp == 0)
		return (CCP_OK);


	/*
	 * Der Vorgang fuer ccp's ist fast gleich mit dem fuer acp's.
	 */

	rc = 0;
	if (pipe(pfd) != 0  ||  pipe(lfd)) {
		syslog(LOG_NOTICE, "-ERR: can't pipe: %m");
		exit (1);
		}
	else if ((pid = fork()) < 0) {
		syslog(LOG_NOTICE, "-ERR: can't fork ccp: %m");
		exit (1);
		}
	else if (pid == 0) {
		int	argc;
		char	*argv[32];

		dup2(pfd[1], 1);	/* stdout nach FTP Client */
		close(pfd[0]);

		dup2(lfd[1], 2);	/* stderr nach syslog */
		close(lfd[0]);

		close(0);
		set_variables(x, cmd, par, ispath);

		copy_string(line, x->config->ccp, sizeof(line));
		argc = split(line, argv, ' ', 30);
		argv[argc] = NULL;
		execvp(argv[0], argv);

		syslog(LOG_NOTICE, "-ERR: can't exec ccp %s: %m", argv[0]);
		exit (1);
		}
	else {
		int	len;

		/*
		 * Nicht gebrauchte fd's schliessen.
		 */

		close(pfd[1]);
		close(lfd[1]);


		/*
		 * syslog Meldung lesen und entsprechende pipe schliessen.
		 */

		*message = 0;
		if ((len = read(lfd[0], message, sizeof(message) - 2)) < 0)
			len = 0;

		message[len] = 0;
		noctrl(message);
		close(lfd[0]);

		if (*message != 0)
			syslog(LOG_NOTICE, "%s", message);



		/*
		 * Fehlermeldung lesen, pipe schliessen.
		 */

		*message = 0;
		if ((len = read(pfd[0], message, sizeof(message) - 2)) < 0)
			len = 0;

		message[len] = 0;
		noctrl(message);
		close(pfd[0]);


		/*
		 * return code holen.
		 */

		if (waitpid(pid, &rc, 0) < 0) {
			syslog(LOG_NOTICE, "-ERR: error while waiting for ccp: %m");
			exit (1);
			}

		rc = WIFEXITED(rc) != 0? WEXITSTATUS(rc): 1;
		if (rc == 0)
			return (CCP_OK);

		if (*message == 0)
			copy_string(message, "553 permission denied", sizeof(message));

/*		snprintf (command, sizeof(command) - 2, "%s%s%s", cmd, (par != 0? " ": ""), par);
 *		syslog(LOG_NOTICE, "ccp: -ERR: %s@%s: %s: %s: rc= %d",
 *				x->username, x->server.name,
 *				command, message, rc);
 */
		}

	if (isdigit(*message))
		cfputs(x, message);
	else {
		snprintf (line, sizeof(line) - 2, "553 %s", message);
		cfputs(x, line);
		}

/*	cfputs(x, "553 permission denied."); */
	return (CCP_ERROR);
}

int call_handler(ftp_t *x, char *cmd, char *par, int ispath)
{
	int	rc, pid, pfd[2];
	char	line[300];

	if (*x->config->handler == 0)
		return (CCP_OK);


	/*
	 * Der Vorgang fuer handler ist fast gleich mit dem fuer ccp's.
	 */

	rc = 0;
	if (pipe(pfd) != 0) {
		syslog(LOG_NOTICE, "-ERR: can't pipe: %m");
		exit (1);
		}
	else if ((pid = fork()) < 0) {
		syslog(LOG_NOTICE, "-ERR: can't fork handler: %m");
		exit (1);
		}
	else if (pid == 0) {
		int	argc;
		char	*argv[32];

		dup2(2, 3);		/* Server stderr sichern */
		dup2(pfd[1], 2);	/* stderr ueber die pipe leiten */
		close(pfd[0]);

		close(1);
		close(0);
		set_variables(x, cmd, par, ispath);

		copy_string(line, x->config->handler, sizeof(line));
		argc = split(line, argv, ' ', 30);
		argv[argc] = NULL;
		execvp(argv[0], argv);

		syslog(LOG_NOTICE, "-ERR: can't exec handler %s: %m", argv[0]);

		/*
		 * Da sich der ftp.server (der parent fork()) nicht wirklich
		 * um den exitcode kuemmert, greifen wir hier zu anderen
		 * Methoden.
		 */

		kill(getppid(), SIGTERM);
		exit (1);
		}
	else {
		int	len;
		char	message[300];

		/*
		 * Nicht gebrauchte fd's schliessen.
		 */

		close(pfd[1]);
/*		close(lfd[1]); */



		/*
		 * Fehlermeldung lesen, pipe schliessen.
		 */

		*message = 0;
		if ((len = read(pfd[0], message, sizeof(message) - 2)) < 0)
			len = 0;

		message[len] = 0;
		noctrl(message);
		close(pfd[0]);

		if (*message != 0)
			syslog(LOG_NOTICE, "%s", message);

		if (waitpid(pid, &rc, 0) < 0) {
			syslog(LOG_NOTICE, "-ERR: error while waiting for ccp: %m");
			exit (1);
			}
		}

	return (0);
}



int run_acp(ftp_t *x)
{
	int	rc, pid, pfd[2];
	char	line[300];
	
	if (*x->config->acp == 0)
		return (0);

	rc = 0;
	if (pipe(pfd) != 0) {
		syslog(LOG_NOTICE, "-ERR: can't pipe: %m");
		exit (1);
		}
	else if ((pid = fork()) < 0) {
		syslog(LOG_NOTICE, "-ERR: can't fork acp: %m");
		exit (1);
		}
	else if (pid == 0) {
		int	argc;
		char	*argv[32], val[80];

		close(0);		/* Das acp kann nicht vom client lesen. */
		dup2(pfd[1], 2);	/* stderr wird vom parent gelesen. */
		close(pfd[0]);

		/*
		 * Umgebungsvariable fuer das acp setzen.
		 */

		setvar(x, "INTERFACE", x->interface);
		snprintf (val, sizeof(val) - 2, "%u", x->port);
		setvar(x, "PORT", val);
		setvar(x, "CLIENT", x->client_ip);
		setvar(x, "CLIENTNAME", x->client);
		setvar(x, "USER", x->username);
		setvar(x, "USERNAME", x->username);

		if (x->config->acp_authenticates != 0)
			setvar(x, "PASSWORD", x->password);

		/*
		 * acp aufrufen.
		 */

		copy_string(line, x->config->acp, sizeof(line));
		argc = split(line, argv, ' ', 30);
		argv[argc] = NULL;
		execvp(argv[0], argv);

		syslog(LOG_NOTICE, "-ERR: can't exec acp %s: %m", argv[0]);
		exit (1);
		}
	else {
		int	len;
		char	message[300];

		close(pfd[1]);
		*message = 0;
		if ((len = read(pfd[0], message, sizeof(message) - 2)) < 0)
			len = 0;

		message[len] = 0;
		noctrl(message);
		close(pfd[0]);

		if (waitpid(pid, &rc, 0) < 0) {
			syslog(LOG_NOTICE, "-ERR: error while waiting for acp: %m");
			exit (1);
			}

		rc = WIFEXITED(rc) != 0? WEXITSTATUS(rc): 1;
		if (*message == 0)
			copy_string(message, rc == 0? "access granted": "access denied", sizeof(message));

		if (rc == 0)
			syslog(LOG_NOTICE, "%s (rc= %d)", message, rc);
		else {
			syslog(LOG_NOTICE, "-ERR: %s (rc= %d)", message, rc);
			exit (1);
			}
		}
		
	return (rc);
}

int dologin(ftp_t *x)
{
	char	*p, word[80], line[300];

	while (1) {
		if (readline_fd(x, 0, line, sizeof(line)) == NULL)
			return (1);

		p = noctrl(line);
		get_word(&p, word, sizeof(word));
		strupr(word);
		if (strcmp(word, "USER") == 0) {
			get_word(&p, x->username, sizeof(x->username));
			
			if (*x->config->denylist != 0) {
				p = x->config->denylist;
				while (*get_quoted(&p, ',', word, sizeof(word)) != 0) {
					if (strcmp(x->username, word) == 0) {
						cfputs(x, "530 access denied");
						syslog(LOG_NOTICE, "-ERR: login denied by configuration: %s", x->username);
						exit (1);
						}
					}
				}

			cfputs(x, "331 password required");
			}
		else if (strcmp(word, "PASS") == 0) {
			if (*x->username == 0) {
				cfputs(x, "503 give USER first");
				continue;
				}
				
			get_word(&p, x->password, sizeof(x->password));
			break;
			}
		else if (strcmp(word, "QUIT") == 0) {
			cfputs(x, "221 goodbye");
			return (2);
			}
		else {
			cfputs(x, "530 login first");
			}
		}


	/*
	 * Now that we have the user login information we call our
	 * access control program if one is given.
	 */

	if (*x->config->acp != 0) {
		if (run_acp(x) != 0) {
			exit (0);
			}
		}


	/*
	 * The acp grants access, let's verify the user login.
	 */

	if (getuid() != 0) {
		syslog(LOG_NOTICE, "-ERR: no access to /etc/passwd, check uid");
		exit (1);
		}
	else {

		/*
		 * Authenticate the user with /etc/shadow or /etc/passwd.
		 */

		unsigned int uid, gid;
		struct passwd *pw;
		
		if (strcmp(x->username, "ftp") == 0  ||  strcmp(x->username, "anonymous") == 0) {
			if (x->config->allow_anon == 0) {
				syslog(LOG_NOTICE, "-ERR: login request for anonamous user denied");
				cfputs(x, "530 access denied");
				exit (1);
				}

			/*
			 * No password checking here, just make sure that we have
			 * the right username for the environment setup.
			 */

			strcpy(x->username, "ftp");
			p = *x->config->anon_user != 0? x->config->anon_user: x->username;
			if ((pw = getpwnam(p)) == NULL) {
				syslog(LOG_NOTICE, "-ERR: can't get anonymous passwd entry: %s", p);
				cfputs(x, "530 access denied");
				exit (1);
				}
			}
		else {
			if (x->config->acp_authenticates != 0)
				syslog(LOG_NOTICE, "user acp authenticated: %s", x->username);
			else {
				char	salt[8], *crypted, *cryptpw;
				struct spwd *spw;

				cryptpw = "";
				if ((spw = getspnam(x->username)) != NULL)
					cryptpw = spw->sp_pwdp;
				else if ((pw = getpwnam(x->username)) != NULL)
					cryptpw = pw->pw_passwd;
				else {
					syslog(LOG_NOTICE, "-ERR: login request for unkown user: %s", x->username);
					cfputs(x, "530 access denied");
					exit (1);
					}

				salt[0] = cryptpw[0];
				salt[1] = cryptpw[1];
				salt[2] = 0;

				crypted = crypt(x->password, salt);
				if (strcmp(cryptpw, crypted) != 0) {
					syslog(LOG_NOTICE, "-ERR: bad login: %s", x->username);
					cfputs(x, "530 bad login");
					exit (1);
					}
				}

			if ((pw = getpwnam(x->username)) == NULL) {
				syslog(LOG_NOTICE, "-ERR: can't get passwd entry: %s", x->username);
				cfputs(x, "530 access denied");
				exit (1);
				}
			}

		if (pw == NULL) {
			syslog(LOG_NOTICE, "-ERR: internal passwd error");
			cfputs(x, "530 access denied");
			exit (1);
			}


		/*
		 * The user's authentication is ok.  Go to the user's home
		 * directory, set his uid and gid.
		 */

		if (*x->config->homedir != 0)
			copy_string(x->home, x->config->homedir, sizeof(x->home));
		else
			copy_string(x->home, pw->pw_dir, sizeof(x->home));

		if (chdir(x->home) != 0) {
			cfputs(x, "500 service unavailable");
			syslog(LOG_NOTICE, "-ERR: can't change to home directory %s, user= %s", x->home, x->username);
			exit (1);
			}


		copy_string(x->gecos, pw->pw_gecos, sizeof(x->gecos));
		if (*x->gecos != 0) {
			char	email[80];

			p = x->gecos;
			while (*get_word(&p, email, sizeof(email)) != 0) {
				if (strchr(email, '@') != NULL) {
					copy_string(x->email, email, sizeof(x->email));
					break;
					}
				}
			}


		copy_string(x->dir, x->home, sizeof(x->dir));
		if (strcmp(x->config->vroot, ".") == 0) {

			/*
			 * Das virtuelle root Verzeichnis ist das Home Verzeichnis
			 * des Benutzers.  Das ist das augenblicklich aktuelle
			 * Verzeichnis.
			 */

			getcwd(x->rootdir, sizeof(x->rootdir));
			/* copy_string(x->rootdir, x->home, sizeof(x->rootdir)); */
			}
		else if (*x->config->vroot != 0) {
			int	len;
			
			copy_string(x->rootdir, x->config->vroot, sizeof(x->rootdir));
			len = strlen(x->rootdir);
			if (strncmp(x->rootdir, x->home, len) != 0  ||  (x->home[len] != '/'  &&  x->home[len] != 0)) {
				syslog(LOG_NOTICE, "-ERR: home outside virtual root: %s, user= %s", x->home, x->username);
				exit (1);
				}
			}
		else if (strstr(x->home, "/./") != NULL) {
			copy_string(x->rootdir, x->home, sizeof(x->rootdir));
			p = strstr(x->rootdir, "/./");
			*p = 0;
			getcwd(x->home, sizeof(x->home));
			}



		uid = x->config->user.uid == 0? pw->pw_uid: x->config->user.uid;
		gid = x->config->user.gid == 0? pw->pw_gid: x->config->user.gid;
		
		if (setregid(gid, gid) != 0) {
			cfputs(x, "500 service unavailable");
			syslog(LOG_NOTICE, "-ERR: can't change gid to %u", gid);
			exit (1);
			}
		else if (setreuid(uid, uid) != 0) {
			cfputs(x, "500 service unavailable");
			syslog(LOG_NOTICE, "-ERR: can't change uid to %u", uid);
			exit (1);
			}
		}

	setuserenv(x);
	cfputs(x, "230 login accepted");
	syslog(LOG_NOTICE, "login accepted: %s", x->username);

	return (0);
}


int apply_sgphack(ftp_t *x, char *par)
{
	unsigned long perms;
	char	dir[300];
	struct stat sbuf;

	if (x->config->sgphack == 0)
		return (0);

	/*
	 * Check if par's directory has the sticky-group-bit set
	 * and is `rwx' for the group (perms = 02x7x).
	 */

	getdir(par, dir, sizeof(dir));
	if (stat(dir, &sbuf) != 0)
		return (0);

	perms = S_IRGRP | S_IWGRP | S_IXGRP | S_ISGID;
	if ((sbuf.st_mode & perms) != perms)
		return (0);

	/*
	 * Ok, make the file read/writable by the owning group.
	 */

	if (stat(par, &sbuf) != 0)
		return (0);

	perms = sbuf.st_mode | S_IRGRP | S_IWGRP;
	chmod(par, perms);

	return (0);
}


void signal_handler(int sig)
{
	syslog(LOG_NOTICE, "-ERR: received signal #%d", sig);
	exit (1);
}

int set_signals(void)
{
	signal(SIGCHLD, SIG_DFL);

	signal(SIGHUP, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGQUIT, signal_handler);
	signal(SIGSEGV, signal_handler);
	signal(SIGPIPE, signal_handler);
	signal(SIGALRM, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGUSR1, signal_handler);
	signal(SIGUSR2, signal_handler);

	return (0);
}


ftpcmd_t *getcmd(char *name)
{
	int	i;

	for (i=0; cmdtab[i].name[0] != 0; i++) {
		if (strcmp(cmdtab[i].name, name) == 0)
			return (&cmdtab[i]);
		}

	return (NULL);
}


int ftp_session(config_t *config)
{
	int	rc;
	char	*p, word[80], par[200], line[300];
	ftpcmd_t *cmd;
	ftp_t	*x;

	set_signals();

	/*
	 * Set socket options to prevent us from the rare case that
	 * we transfer data to/from the client before the client has
	 * seen our "150 ..." message.
	 * Seems so that is doesn't work on all systems.
	 * So temporary only enable it on linux. 
	 */

#if defined(__linux__)

	rc = 1;
	if (setsockopt(1, SOL_TCP, TCP_NODELAY, &rc, sizeof(rc)) != 0)
		syslog(LOG_NOTICE, "can't set TCP_NODELAY, error= %s", strerror(errno));

#endif

	x = allocate(sizeof(ftp_t));
	x->config = config;

	/*
	 * Necessary initialisation to avoid closing stdin on error in first
	 * DTC request.
	 */

	x->ch.local = -1;
	x->ch.data  = -1;


	if (get_client_info(x, 0) < 0) {
		syslog(LOG_NOTICE, "-ERR: can't get client info: %m");
		exit (1);
		}

	x->port = get_interface_info(0, x->interface, sizeof(x->interface));
	syslog(LOG_NOTICE, "connected to client: %s", x->client);

	cfputs(x, "220 server ready - login please");
	if ((rc = dologin(x)) < 0)
		return (1);
	else if (rc == 2)
		return (0);

	call_handler(x, "+LOGIN", "", 0);
	while ((p = cfgets(x, line, sizeof(line))) != NULL) {
		if (*p == '\001') {
			int	rc;

			if (*x->ch.command != 0) {
				syslog(LOG_NOTICE, "%s %s %s %s %ld bytes",
					x->username, x->client,
					x->ch.command, x->ch.filename, x->ch.bytes);
				}

			rc = 0;
			if (strcmp(x->ch.command, "RETR") == 0)
				rc = call_handler(x, "+RETR", x->ch.filename, 1);
			else if (strcmp(x->ch.command, "STOR") == 0)
				rc = call_handler(x, "+STOR", x->ch.filename, 1);
			else if (strcmp(x->ch.command, "STOU") == 0)
				rc = call_handler(x, "+STOU", x->ch.filename, 1);
			else if (strcmp(x->ch.command, "APPE") == 0)
				rc = call_handler(x, "+APPE", x->ch.filename, 1);

			if (1  ||  rc == 0) {
				cfputs(x, "226 transfer complete");
				}
			else {
				syslog(LOG_NOTICE, "-ERR: unexpected handler exit code");
				exit (1);
				}

			*x->ch.command  = 0;
			*x->ch.filename = 0;
			continue;
			}


		p = noctrl(line);
		get_word(&p, word, sizeof(word));
		strupr(word);
		if ((cmd = getcmd(word)) == NULL  ||  cmd->par == -1) {
			cfputs(x, "500 command not implemented");
			syslog(LOG_NOTICE, "command not implemented: %s", word);
			continue;
			}


		/*
		 * Befehlsparameter holen ...
		 */

		*x->origpar = 0;
		*par = 0;
		if (cmd->par > 0) {
			if (strcmp(cmd->name, "SITE") == 0) {
				int	c, i;
				char	*r;

				if (*x->config->sitehandler == 0) {
					cfputs(x, "502 command not implemented");
					continue;
					}

				copy_string(x->sitepar, p, sizeof(x->sitepar));
				if (*x->sitepar == 0) {
					cfputs(x, "501 parameter required");
					continue;
					}

				for (i=0; (c = x->sitepar[i]) != 0; i++) {
					if (c <= ' ')
						break;

					if (islower(c))
						x->sitepar[i] = toupper(c);
					}
					
				if ((r = strrchr(x->sitepar, ' ')) != NULL)
					get_word(&r, par, sizeof(par));
				else
					copy_string(par, x->sitepar, sizeof(par));
				}
			else if (strcmp(cmd->name, "CDUP") == 0) {

				/*
				 * Parameter fuer virtual root directory
				 * Test setzen.
				 */

				strcpy(par, "..");
				}
			else {
				int	c;

				p = skip_ws(p);
				if (strcmp(cmd->name, "LIST") == 0  ||  strcmp(cmd->name, "NLIST") == 0) {
					char	skip[80];

					/*
					 * Das LIST Kommando unterstuetzt keine Optionen.
					 * Die RFC erwaehnt auch keine.  Also was soll's,
					 * wenn welche das sind ignorieren wir sie einfach.
					 */

					while (*(p = skip_ws(p)) == '-')
						get_word(&p, skip, sizeof(skip));
					}


				if (x->config->allow_blanks == 0)
					get_word(&p, par, sizeof(par));
				else {
					if (((c = *p) == '\''  ||  c == '"')  &&
					    x->config->allow_quotes) {
					    	p++;
						get_quoted(&p, c, par, sizeof(par));
						}
					else
						copy_string(par, p, sizeof(par));
					}
				}


			/*
			 * ... und, falls es sich um einen Datei oder
			 * Verzeichnisnamen handelt, in einen absoluten
			 * Pfad umwandeln.
			 */

			if (cmd->ispath != 0) {

				/*
				 * We save the original command parameter just in case
				 * that we need it later.  For NLST we do.
				 */

				copy_string(x->origpar, par, sizeof(x->origpar));


				if (strcmp(cmd->name, "LIST") == 0  ||  strcmp(cmd->name, "NLST")) {
					if (*par == 0)
						strcpy(par, "./*");
					}
					
				if (debug != 0)
					fprintf (stderr, "0: %s, path= %s\n", word, par);

				if (normalize_path(x, par, sizeof(par)) != 0) {
					cfputs(x, "550 no such file or directory");
					syslog(LOG_NOTICE, "request to access %s denied", par);
					continue;
					}

				if (debug != 0)
					fprintf (stderr, "1: %s, path= %s\n", word, par);

				if (*get_syspath(x, par, sizeof(par)) == 0) {
					cfputs(x, "550 no such file or directory");
					syslog(LOG_NOTICE, "internal error: syspath %s", par);
					continue;
					}

				if (debug != 0)
					fprintf (stderr, "2: %s, path= %s\n", word, par);
				}
			}

		/*
		 * command control program aufrufen.
		 */

		if (cmd->useccp != 0) {
			if (run_ccp(x, word, par, cmd->ispath) != CCP_OK)
				continue;
			}


		/*
		 * Ausfuehren des Befehls.
		 */

		if (strcmp(cmd->name, "QUIT") == 0) {
			doquit(x);
			break;
			}
		else if (strcmp(cmd->name, "PORT") == 0)
			doport(x, word, par);
		else if (strcmp(cmd->name, "PASV") == 0)
			dopasv(x);
		else if (strcmp(cmd->name, "SYST") == 0)
			cfputs(x, "215 UNIX Type: L8");
		else if (strcmp(cmd->name, "TYPE") == 0) {
			if (*par != 'I'  &&  *par != 'A')
				cfputs(x, "500 illegal type");
			else {
				x->type = (*par == 'I')? TYPE_BIN: TYPE_ASC;
				cfputs(x, "200 transfer type accepted.");
				}
			}

		else if (strcmp(cmd->name, "CWD") == 0) {
			if (chdir(par) != 0)
				cfpute(x, "550");
			else
				cfputs(x, "250 ok");
			}
		else if (strcmp(cmd->name, "CDUP") == 0) {
			if (chdir(par) != 0)
				cfpute(x, "550");
			else
				cfputs(x, "250 ok");
			}

		else if (strcmp(cmd->name, "PWD") == 0) {
			char	dir[200];

			if (getcwd(dir, sizeof(dir)) == NULL)
				cfputs(x, "550 not available");
			else {
				get_ftppath(x, dir, sizeof(dir));
				snprintf (line, sizeof(line) - 2, "257 \"%s\" is current directory", dir);
				cfputs(x, line);
				}
			}

		else if (strcmp(cmd->name, "LIST") == 0  ||  strcmp(cmd->name, "NLST") == 0) {
			int	fd;

			fd = listdir(x, par, strcmp(cmd->name, "LIST") == 0? 0: 1);
			if (fd >= 0) {
				cfputs(x, "150 listing follows");
				connect_ch(x, &x->ch, OP_GET, fd, 1, TYPE_ASC, "", "");
				}
			else {
				cfputs(x, "500 service unavailable");
				}
			}

		else if (strcmp(cmd->name, "ABOR") == 0) {
			if (x->ch.state == PORT_UNCONNECTED)
				cfputs(x, "503 no data transfer");
			else {
				close_ch(x, &x->ch);
				syslog(LOG_NOTICE, "transfer aborted: %s %s: %ld bytes", x->ch.command, x->ch.filename, x->ch.bytes);

				if (*x->ch.filename != 0) {
					if (strcmp(x->ch.command, "RETR") == 0)
						call_handler(x, "RETR-", x->ch.filename, 1);
					else if (strcmp(x->ch.command, "STOR") == 0)
						call_handler(x, "STOR-", x->ch.filename, 1);
					else if (strcmp(x->ch.command, "STOU") == 0)
						call_handler(x, "STOU-", x->ch.filename, 1);
					else if (strcmp(x->ch.command, "APPE") == 0)
						call_handler(x, "APPE-", x->ch.filename, 1);
					}

				cfputs(x, "226, transfer aborted");
				}

			*x->ch.command  = 0;
			*x->ch.filename = 0;
			}

		else if (strcmp(cmd->name, "REST") == 0) {
			x->restart = strtoul(par, NULL, 10);
			cfputs(x, "350 ok");
			}

		else if (strcmp(cmd->name, "MDTM") == 0  ||  strcmp(cmd->name, "SIZE") == 0) {
			struct stat sbuf;

			if (*par == 0) {
				cfputs(x, "501 missing parameter");
				continue;
				}
			else if (stat(par, &sbuf) != 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "can't stat file: %s, error= %s", par, strerror(errno));
				continue;
				}
			else if (S_ISREG(sbuf.st_mode) == 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "not a plain file: %s", par);
				continue;
				}

			if (strcmp(cmd->name, "SIZE") == 0) {
				snprintf (line, sizeof(line) - 2, "213 %ld", sbuf.st_size);
				cfputs(x, line);
				}
			else if (strcmp(cmd->name, "MDTM") == 0) {
				char	buffer[80];
				struct tm *date;

				date = gmtime(&sbuf.st_mtime);
				strftime(buffer, sizeof(buffer), "%Y%m%d%H%M%S", date);
				snprintf (line, sizeof(line) - 2, "213 %s", buffer);
				cfputs(x, line);
				}
			}

		else if (strcmp(cmd->name, "RETR") == 0) {
			int	fd;
			struct stat sbuf;

			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if (stat(par, &sbuf) != 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "can't stat file: %s, error= %s", par, strerror(errno));
				}
			else if (S_ISREG(sbuf.st_mode) == 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "not a plain file: %s", par);
				}
			else if ((fd = open(par, O_RDONLY)) < 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "can't open file: %s, error= %m", par);
				}
			else {
				if (x->restart != 0) {
					if (lseek(fd, x->restart, SEEK_SET) < 0)
						syslog(LOG_NOTICE, "restart error: %m");
					}

				cfputs(x, "150 data follows");
				connect_ch(x, &x->ch, OP_GET, fd, 0, x->type, "RETR", par);
				call_handler(x, "RETR", x->ch.filename, 1);
				}
			}

		else if (strcmp(cmd->name, "STOR") == 0) {
			int	fd;

			fd = -1;
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else {
				if (x->restart == 0)
					fd = open(par, O_WRONLY | O_CREAT | O_TRUNC, 0666);
				else {
					if ((fd = open(par, O_WRONLY | O_CREAT, 0666)) >= 0) {
						if (lseek(fd, x->restart, SEEK_SET) < 0)
							syslog(LOG_NOTICE, "restart error: %m");
						}
					}

				if (fd < 0) {
					cfputs(x, "550 no access");
					syslog(LOG_NOTICE, "can't write file: %s, error= %m", par);
					}
				else {
					if (x->config->sgphack != 0)
						apply_sgphack(x, par);

					cfputs(x, "150 send data");
					connect_ch(x, &x->ch, OP_PUT, fd, 0, x->type, "STOR", par);
					call_handler(x, "STOR", x->ch.filename, 1);
					}
				}
			}

		else if (strcmp(cmd->name, "STOU") == 0) {
			int	fd;
			char	filename[200], uniquename[80], cwd[200];

			getcwd(cwd, sizeof(cwd));
			if (*x->config->unamehandler == 0) {
				cfputs(x, "550 access denied");
				syslog(LOG_NOTICE, "no uniquename handler configured: STOU");
				}
			else if (getuniquename(x, uniquename, sizeof(uniquename)) != 0  ||  *uniquename == 0) {
				cfputs(x, "550 access denied");
				syslog(LOG_NOTICE, "STOU access denied: %s", cwd);
				}
			else {
				snprintf (filename, sizeof(filename) - 2, "%s/%s", cwd, uniquename);
				if ((fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0666)) < 0) {
					close_ch(x, &x->ch);
					cfputs(x, "550 access denied");
					syslog(LOG_NOTICE, "can't write/create file: %s, error= %m", filename);
					}
				else {
					char	response[200];

					if (x->config->sgphack != 0)
						apply_sgphack(x, par);

					snprintf (response, sizeof(response) - 2, "150 %s ok, send data", uniquename);
					cfputs(x, response);
					connect_ch(x, &x->ch, OP_PUT, fd, 0, x->type, "STOU", filename);
					call_handler(x, "STOU", x->ch.filename, 1);
					}
				}
			}

		else if (strcmp(cmd->name, "APPE") == 0) {
			int	fd;

			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if ((fd = open(par, O_WRONLY | O_CREAT | O_APPEND, 0666)) < 0) {
				cfputs(x, "550 no access");
				syslog(LOG_NOTICE, "can't write file: %s, error= %m", par);
				}
			else {
				if (x->config->sgphack != 0)
					apply_sgphack(x, par);

				cfputs(x, "150 send data");
				connect_ch(x, &x->ch, OP_PUT, fd, 0, x->type, "STOR", par);
				call_handler(x, "APPE", x->ch.filename, 1);
				}
			}

		else if (strcmp(cmd->name, "DELE") == 0) {
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if (unlink(par) != 0)
				cfputs(x, "550 access denied");
			else
				cfputs(x, "250 file deleted");
			}

		else if (strcmp(cmd->name, "RNFR") == 0) {
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else {
				copy_string(x->renamefile, par, sizeof(x->renamefile));
				cfputs(x, "350 send new filename");
				}
			}
		else if (strcmp(cmd->name, "RNTO") == 0) {
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if (*x->renamefile == 0)
				cfputs(x, "500 give source filename first");
			else if (rename(x->renamefile, par) != 0)
				cfputs(x, "550 access denied");
			else {
				if (x->config->sgphack != 0)
					apply_sgphack(x, par);

				cfputs(x, "250 file renamed");
				}
			}

		else if (strcmp(cmd->name, "MKD") == 0) {
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if (mkdir(par, 0777) != 0)
				cfpute(x, "550");
			else {
				if (x->config->sgphack != 0)
					apply_sgphack(x, par);

				cfputs(x, "257 directory created");
				}
			}

		else if (strcmp(cmd->name, "RMD") == 0) {
			if (*par == 0)
				cfputs(x, "501 missing parameter");
			else if (rmdir(par) != 0)
				cfpute(x, "550");
			else
				cfputs(x, "250 directory deleted");
			}

		else if (strcmp(cmd->name, "NOOP") == 0) {
			cfputs(x, "200 ok");
			}

		else if (strcmp(cmd->name, "SITE") == 0)
			dosite(x, par);

		else {
			syslog(LOG_NOTICE, "unknown command: %s", word);
			cfputs(x, "502 unknown command");
			}

		if (strcmp(cmd->name, "RNFR") != 0)
			*x->renamefile = 0;

		if (strcmp(cmd->name, "STOR") == 0  ||  strcmp(cmd->name, "RETR") == 0  ||
		    strcmp(cmd->name, "DELE") == 0  ||  strcmp(cmd->name, "MKD") == 0  ||
		    strcmp(cmd->name, "RMD") == 0)
			x->restart = 0;

		if (cmd->usehandler == H_POST)
			call_handler(x, word, par, cmd->ispath);
		}

	syslog(LOG_NOTICE, "+OK: ftpserver terminating");
	return (0);
}



